from eralchemy.models import sanitize_mermaid


def test_mermaid_escape():
    assert "0" == sanitize_mermaid("0")
    assert "_0" == sanitize_mermaid("0", is_er=True)
    assert "_-" == sanitize_mermaid("-", is_er=True)
    assert "-" == sanitize_mermaid("-")
    assert "left_right" == sanitize_mermaid("left/right")
    assert "test" == sanitize_mermaid("test")
    assert "test7" == sanitize_mermaid("test7", is_er=True)
    assert "public_table_column" == sanitize_mermaid("public.table.column")
    assert "public_table_column" == sanitize_mermaid("public%table!column")



from sqlalchemy import Column, ForeignKey, String
from sqlalchemy.orm import declarative_base

from eralchemy.models import Relation, Table
from eralchemy.sqla import (
    declarative_to_intermediary,
)


def test_columns_parent():
    Base = declarative_base()

    class Parent(Base):
        __tablename__ = "parent"
        id = Column(String(), primary_key=True)

    class Child(Base):
        __tablename__ = "child"
        id = Column(String(), primary_key=True)
        parent = Column(String(), ForeignKey(Parent.id), primary_key=True)

    tables, relationships = declarative_to_intermediary(Base)

    assert len(tables) == 2
    assert len(relationships) == 1
    assert all(isinstance(t, Table) for t in tables)
    assert all(isinstance(r, Relation) for r in relationships)
    relation = relationships[0]
    assert relation.right_cardinality == "*"
    assert relation.left_cardinality == "1"


def test_columns_one_to_many_parent():
    Base = declarative_base()

    class Parent(Base):
        __tablename__ = "parent"
        id = Column(String(), primary_key=True)

    class Child(Base):
        __tablename__ = "child"
        id = Column(String(), primary_key=True)
        parent = Column(String(), ForeignKey(Parent.id))

    tables, relationships = declarative_to_intermediary(Base)

    assert len(tables) == 2
    assert len(relationships) == 1
    assert all(isinstance(t, Table) for t in tables)
    assert all(isinstance(r, Relation) for r in relationships)
    relation = relationships[0]
    assert relation.right_cardinality == "*"
    assert relation.left_cardinality == "?"


def test_columns_one_to_one_parent():
    Base = declarative_base()

    class Parent(Base):
        __tablename__ = "parent"
        id = Column(String(), primary_key=True)

    class Child(Base):
        __tablename__ = "child"
        id = Column(String(), ForeignKey(Parent.id), primary_key=True)

    tables, relationships = declarative_to_intermediary(Base)

    assert len(tables) == 2
    assert len(relationships) == 1
    assert all(isinstance(t, Table) for t in tables)
    assert all(isinstance(r, Relation) for r in relationships)
    relation = relationships[0]
    assert relation.right_cardinality == "1"
    assert relation.left_cardinality == "1"


def test_compound_one_to_one_parent():
    Base = declarative_base()

    class Parent(Base):
        __tablename__ = "parent"
        id = Column(String(), primary_key=True)
        id2 = Column(String(), primary_key=True)

    class Child(Base):
        __tablename__ = "child"
        id = Column(String(), ForeignKey(Parent.id), primary_key=True)
        id2 = Column(String(), ForeignKey(Parent.id2), primary_key=True)

    tables, relationships = declarative_to_intermediary(Base)

    assert len(tables) == 2
    assert len(relationships) == 2
    assert all(isinstance(t, Table) for t in tables)
    assert all(isinstance(r, Relation) for r in relationships)
    for relation in relationships:
        assert relation.right_cardinality == "1"
        assert relation.left_cardinality == "1"



import argparse
import base64
import copy
import logging
import os
import re
import sys
import typing
from functools import partial
from importlib.metadata import PackageNotFoundError, version

from sqlalchemy.engine.url import make_url
from sqlalchemy.exc import ArgumentError

from .cst import config
from .helpers import check_args, plantuml_convert
from .parser import (
    ParsingException,
    line_iterator_to_intermediary,
    markdown_file_to_intermediary,
)
from .sqla import (
    database_to_intermediary,
    declarative_to_intermediary,
    metadata_to_intermediary,
)

USE_PYGRAPHVIZ = True
GRAPHVIZ_AVAILABLE = True
try:
    from pygraphviz.agraph import AGraph

    logging.debug("using pygraphviz")
except ImportError:
    USE_PYGRAPHVIZ = False
    try:
        from graphviz import Source

        logging.debug("using graphviz")
    except ImportError:
        logging.error("either pygraphviz or graphviz should be installed")
        GRAPHVIZ_AVAILABLE = False

try:
    __version__ = version(__package__)
except PackageNotFoundError:
    __version__ = "na"


def cli(args=None) -> None:
    """Entry point for the application script."""
    parser = get_argparser()

    args = parser.parse_args(args)
    check_args(args)
    if args.v:
        print(f"eralchemy version {__version__}.")
        exit(0)
    output = render_er(
        args.i,
        args.o,
        args.m or "auto",
        title=args.title,
        include_tables=args.include_tables,
        include_columns=args.include_columns,
        exclude_tables=args.exclude_tables,
        exclude_columns=args.exclude_columns,
        schema=args.s,
    )
    if output:
        with os.fdopen(sys.stdout.fileno(), "wb", closefd=False) as stdout:
            stdout.write(output)
            stdout.flush()


def get_argparser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="eralchemy")
    parser.add_argument("-i", nargs="?", help="Database URI to process.")
    parser.add_argument("-o", nargs="?", help="Name of the file to write.")
    parser.add_argument("-s", nargs="?", help="Name of the schema.")
    parser.add_argument("--title", nargs="?", help="Add a title to the output graph")
    parser.add_argument(
        "-m",
        nargs="?",
        help="Output mode to write format, default: auto",
    )
    parser.add_argument(
        "--exclude-tables",
        "-x",
        nargs="+",
        help="Name of tables not to be displayed.",
    )
    parser.add_argument(
        "--exclude-columns",
        nargs="+",
        help="Name of columns not to be displayed (for all tables).",
    )
    parser.add_argument(
        "--include-tables",
        nargs="+",
        help="Name of tables to be displayed alone.",
    )
    parser.add_argument(
        "--include-columns",
        nargs="+",
        help="Name of columns to be displayed alone (for all tables).",
    )
    parser.add_argument("-v", help="Prints version number.", action="store_true")
    return parser


def intermediary_to_markdown(tables, relationships, title=""):
    """Saves the intermediary representation to markdown."""
    er_markup = _intermediary_to_markdown(tables, relationships)
    if title:
        format_title = config["MARKDOWN_TITLE"].format(title)
        er_markup_with_config = f"{format_title}\n{er_markup}"
    else:
        er_markup_with_config = er_markup
    return er_markup_with_config.encode()


def intermediary_to_mermaid(tables, relationships, title=""):
    """Saves the intermediary representation to markdown."""
    markup = _intermediary_to_mermaid(tables, relationships)
    if title:
        markup = f"""---
title: {title}
---
{markup}
"""
    md_markup = f"<!--\n\n{markup}\n\n-->\n"
    markup_b64 = base64.urlsafe_b64encode(markup.encode("utf8")).decode("ascii")
    md_markup += f"![](https://mermaid.ink/img/{markup_b64})\n"
    return md_markup.encode()


def intermediary_to_mermaid_er(tables, relationships, title=""):
    """Saves the intermediary representation to markdown."""
    markup = _intermediary_to_mermaid_er(tables, relationships)
    if title:
        markup = f"""---
title: {title}
---
{markup}
"""
    md_markup = f"<!--\n\n{markup}\n\n-->\n"
    markup_b64 = base64.urlsafe_b64encode(markup.encode("utf8")).decode("ascii")
    md_markup += f"![](https://mermaid.ink/img/{markup_b64})\n"
    return md_markup.encode()


def intermediary_to_dot(tables, relationships, title=""):
    """Save the intermediary representation to dot format."""
    dot_file = _intermediary_to_dot(tables, relationships, title)
    return dot_file.encode()


def intermediary_to_schema(tables, relationships, title="", extension="png"):
    """Transforms and save the intermediary representation to the file chosen."""
    dot_file = _intermediary_to_dot(tables, relationships, title)
    if not GRAPHVIZ_AVAILABLE:
        raise Exception("either pygraphviz or graphviz should be installed")
    if USE_PYGRAPHVIZ:
        graph = AGraph()
        graph = graph.from_string(dot_file)
        return graph.draw(prog="dot", format=extension)
    else:
        graph = Source(dot_file, engine="dot")
        return graph.pipe(format=extension)


def intermediary_to_puml(tables, relationships, output, title=""):
    """Saves the intermediary representation to PlantUML."""
    puml_markup = _intermediary_to_puml(tables, relationships)
    puml_markup = "\n".join(
        (
            "left to right direction",
            f"title {title}\n {puml_markup}" if title else "",
            puml_markup,
        )
    )
    markup_encoded = plantuml_convert(puml_markup)
    puml_markup += (
        f"\nfooter [[https://www.plantuml.com/plantuml/svg/{markup_encoded}"
        "{link to PlantUML server} Link to PlantUML server]]"
    )

    puml_markup = f"@startuml\n{puml_markup}\n@enduml"
    return puml_markup.encode()


def _intermediary_to_markdown(tables, relationships):
    """Returns the er markup source in a string."""
    t = "\n".join(t.to_markdown() for t in tables)
    r = "\n".join(r.to_markdown() for r in relationships)
    return f"{t}\n{r}"


def _intermediary_to_mermaid(tables, relationships):
    """Returns the er markup source in a string."""
    t = "\n".join(t.to_mermaid() for t in tables)
    r = "\n".join(r.to_mermaid() for r in relationships)
    return f"classDiagram\n{t}\n{r}"


def _intermediary_to_mermaid_er(tables, relationships):
    """Returns the er markup source in a string."""
    t = "\n".join(t.to_mermaid_er() for t in tables)
    r = "\n".join(r.to_mermaid_er() for r in relationships)
    return f"erDiagram\n{t}\n{r}"


def _intermediary_to_dot(tables, relationships, title=""):
    """Returns the dot source representing the database in a string."""
    t = "\n".join(t.to_dot() for t in tables)
    r = "\n".join(r.to_dot() for r in relationships)

    graph_config = (
        f"""{config["DOT_GRAPH_BEGINNING"]}
         label="{title}"
         labelloc=t\n"""
        if title
        else config["DOT_GRAPH_BEGINNING"]
    )
    return f"{graph_config}\n{t}\n{r}\n}}"


def _intermediary_to_puml(tables, relationships):
    """Returns the er markup source in a string."""
    t = "\n".join(t.to_puml() for t in tables)
    r = "\n".join(r.to_puml() for r in relationships)
    return f"{t}\n{r}"


# Routes from the class name to the function transforming this class in
# the intermediary representation.
switch_input_class_to_method = {
    "MetaData": metadata_to_intermediary,
    "DeclarativeMeta": declarative_to_intermediary,
    # For compatibility with Flask-SQLAlchemy
    "_BoundDeclarativeMeta": declarative_to_intermediary,
    # Renamed in Flask-SQLAlchemy 2.3
    "DefaultMeta": declarative_to_intermediary,
    "DeclarativeAttributeIntercept": declarative_to_intermediary,
}

# Routes from the mode to the method to transform the intermediary
#  representation to the desired output.
switch_output_mode_auto = {
    "er": intermediary_to_markdown,
    "mermaid": intermediary_to_mermaid,
    "mermaid_er": intermediary_to_mermaid_er,
    "graph": intermediary_to_schema,
    "graph_svg": partial(intermediary_to_schema, extension="svg"),
    "dot": intermediary_to_dot,
    "puml": intermediary_to_puml,
    "pu": intermediary_to_puml,
}

# Routes from the file extension to the method to transform
# the intermediary representation to the desired output.
switch_output_mode = {
    "er": intermediary_to_markdown,
    "md": intermediary_to_mermaid,
    "dot": intermediary_to_dot,
    "puml": intermediary_to_puml,
}


def all_to_intermediary(filename_or_input, schema=None):
    """Dispatch the filename_or_input to the different function to produce the intermediary syntax.

    All the supported classes names are in `switch_input_class_to_method`.
    The input can also be a list of strings in markdown format or a filename finishing by '.er' containing markdown
    format.
    """
    # Try to convert from the name of the class
    input_class_name = filename_or_input.__class__.__name__
    try:
        this_to_intermediary = switch_input_class_to_method[input_class_name]
        tables, relationships = this_to_intermediary(filename_or_input)
        return tables, relationships
    except KeyError:
        pass

    # try to read markdown file.
    if isinstance(filename_or_input, str):
        if filename_or_input.split(".")[-1] == "er":
            return markdown_file_to_intermediary(filename_or_input)

    # try to read a markdown in a string
    if not isinstance(filename_or_input, str):
        if all(isinstance(e, str) for e in filename_or_input):
            return line_iterator_to_intermediary(filename_or_input)

    # try to read DB URI might raise ArgumentError.
    try:
        make_url(filename_or_input)
        return database_to_intermediary(filename_or_input, schema=schema)
    except ArgumentError as e:
        raise ValueError(f"Cannot process filename_or_input {input_class_name}: {e}")


def get_output_mode(output: typing.Union[str, None], mode: str):
    """From the output name and the mode returns a the function that will transform the intermediary representation to the output."""
    if mode != "auto":
        try:
            return switch_output_mode_auto[mode]
        except KeyError:
            raise ValueError(f'Mode "{mode}" is not supported.')

    if output is None:
        raise ValueError("Mode must be set if output file is None")

    extension = output.split(".")[-1]
    try:
        return switch_output_mode[extension]
    except KeyError:
        return partial(intermediary_to_schema, extension=extension)


def filter_resources(
    tables,
    relationships,
    include_tables=None,
    include_columns=None,
    exclude_tables=None,
    exclude_columns=None,
):
    """Filter the resources.

    Include the following:
        1. Tables and relationships with tables present in the include_tables (lst of str, tables names)
        2. Columns (of whichever table) present in the include_columns (lst of str, columns names)
    Exclude the following:
        1. Tables and relationships with tables present in the exclude_tables (lst of str, tables names)
        2. Columns (of whichever table) present in the exclude_columns (lst of str, columns names)
    Disclosure note:
        All relationships are taken into consideration before ignoring columns.
        In other words, if one excludes primary or foreign keys, it will still keep the relations display amongst tables.
    """
    _tables = copy.deepcopy(tables)
    _relationships = copy.deepcopy(relationships)

    include_tables_re = re.compile(
        "|".join(f"({name})" for name in (include_tables or [t.name for t in _tables])),
    )
    include_columns_re = re.compile(
        "|".join(
            f"({name})"
            for name in (include_columns or [c.name for t in _tables for c in t.columns])
        ),
    )
    exclude_tables_re = re.compile(
        "|".join(f"({name})" for name in (exclude_tables or [])),
    )
    exclude_columns_re = re.compile(
        "|".join(f"({name})" for name in (exclude_columns or [])),
    )

    def check_table(name):
        return not exclude_tables_re.fullmatch(name) and include_tables_re.fullmatch(
            name,
        )

    _tables = [t for t in _tables if check_table(t.name)]
    _relationships = [
        r
        for r in _relationships
        if not exclude_tables_re.fullmatch(r.right_table)
        and not exclude_tables_re.fullmatch(r.left_table)
        and include_tables_re.fullmatch(r.right_table)
        and include_tables_re.fullmatch(r.left_table)
    ]

    def check_column(name):
        return not exclude_columns_re.fullmatch(name) and include_columns_re.fullmatch(
            name,
        )

    for t in _tables:
        t.columns = sorted([c for c in t.columns if check_column(c.name)])

    return _tables, _relationships


def render_er(
    input,
    output: typing.Union[str, None],
    mode="auto",
    include_tables=None,
    include_columns=None,
    exclude_tables=None,
    exclude_columns=None,
    schema=None,
    title=None,
):
    """Transform the metadata into a representation.

    :param input: Possible inputs are instances of:
        MetaData: SQLAlchemy Metadata
        DeclarativeMeta: SQLAlchemy declarative Base
    :param output: name of the file to output the rendered graph to.
        Returns text as str if set to None
    :param mode: str in list:
        'er': writes to a file the markup to generate an ER style diagram.
        'graph': writes the image of the ER diagram.
        'mermaid': writes to a file the markup to generate an Mermaid-JS style diagram
        'mermaid_er': writes the markup to generate an Mermaid-JS ER diagram
        'dot': write to file the diagram in dot format.
        'auto': choose from the filename:
            '*.er': writes to a file the markup to generate an ER style diagram.
            '.dot': returns the graph in the dot syntax.
            '.md': writes to a file the markup to generate an Mermaid-JS style diagram
            else: return a graph to the format graph
    :param include_tables: lst of str, table names to include, None means include all
    :param include_columns: lst of str, column names to include, None means include all
    :param exclude_tables: lst of str, table names to exclude, None means exclude nothing
    :param exclude_columns: lst of str, field names to exclude, None means exclude nothing
    :param schema: name of the schema
    :param title: title of the graph, only for .er, .dot, .png, .jpg outputs.
    """
    try:
        tables, relationships = all_to_intermediary(input, schema=schema)
        tables, relationships = filter_resources(
            tables,
            relationships,
            include_tables=include_tables,
            include_columns=include_columns,
            exclude_tables=exclude_tables,
            exclude_columns=exclude_columns,
        )
        intermediary_to_output = get_output_mode(output, mode)
        text = intermediary_to_output(tables, relationships, title)
        # graphviz does not yet support printing to stdout
        # but writes directly to the output file
        # https://github.com/xflr6/graphviz/pull/234
        if output is None:
            return text
        with open(output, "wb") as file_out:
            file_out.write(text)
    except ImportError as e:
        module_name = str(e).split()[-1]
        print(f'Please install {module_name} using "pip install {module_name}".', file=sys.stderr)
    except (FileNotFoundError, ValueError) as e:
        print(f"rendering failed: {e}", file=sys.stderr)
    except ParsingException as e:
        print(f"render failed: {e}", file=sys.stderr)


if __name__ == "__main__":
    # cli("-i example/forum.er -o test.dot".split(" "))
    config["DOT_GRAPH_BEGINNING"] = """graph {
    graph [rankdir=TD];
    node [label="\\N",
        shape=plaintext
    ];
    edge [color=gray50,
        minlen=2,
        style=dashed
    ];"""
    cli()



"""All the intermediary syntax.

We can several kinds of models can be translated to this syntax.
"""

from __future__ import annotations

import operator
import re
from abc import ABC, abstractmethod
from typing import ClassVar

from .cst import config


class Drawable(ABC):
    """Abstract class to represent all the objects which are drawable."""

    RE: ClassVar[re.Pattern[str]]

    def to_markdown(self) -> str:
        """Transforms the intermediary object to its syntax in the er markup."""
        raise NotImplementedError()

    def to_dot(self) -> str:
        """Transforms the intermediary object to its syntax in the dot format."""
        raise NotImplementedError()

    def to_puml(self) -> str:
        """Transforms the intermediary object to it's syntax in the PlantUML format."""
        raise NotImplementedError()

    def __eq__(self, other) -> bool:
        return self.__dict__ == other.__dict__

    @staticmethod
    @abstractmethod
    def make_from_match(match: re.Match) -> Drawable:
        """Used in the parsing of files. Transforms a regex match to a Drawable object."""

    def __str__(self) -> str:
        return self.to_markdown()


def sanitize_mermaid(text: str, *, is_er: bool = False):
    RE = re.compile("[^0-9a-zA-Z_-]+")
    """Mermaid does not allow special characters in column names"""
    if not text:
        return text
    if is_er and (text[0].isdigit() or text[0] == "-"):
        # mermaid ER does not allow leading dash or digits in column names
        text = "_" + text
    return re.sub(RE, "_", text)


class Column(Drawable):
    """Represents a Column in the intermediaty syntax."""

    RE = re.compile(
        r'(?P<primary>\*?)(?P<name>\w+(\s*\w+)*)\s*(\{label:\s*"(?P<label>[^"]+)"\})?',
    )

    @staticmethod
    def make_from_match(match: re.Match) -> Column:
        return Column(
            name=match.group("name"),
            type=match.group("label"),
            is_key="*" in match.group("primary"),
            is_null="*" not in match.group("primary"),
        )

    def __init__(self, name: str, type=None, is_key: bool = False, is_null=None):
        """Initialize the Column class.

        :param name: (str) Name of the column
        :param type:
        :param is_key:
        :param is_null:
        :return:
        """
        self.name = name
        self.type = type
        self.is_key = is_key
        if is_null is None:
            self.is_null = not is_key
        else:
            self.is_null = is_null

    def __lt__(self, other: Column) -> bool:
        if self.is_key > other.is_key:
            return True
        elif self.is_key < other.is_key:
            return False
        else:
            return self.name < other.name

    @property
    def key_symbol(self) -> str:
        return "*" if self.is_key else ""

    def to_markdown(self) -> str:
        name = sanitize_mermaid(self.name)
        return f'    {self.key_symbol}{name} {{label:"{self.type}"}}'

    def to_mermaid(self) -> str:
        return " {}{} {}{}".format(
            self.key_symbol,
            self.type.replace("(", "<").replace(")", ">"),
            self.name,
            " NOT NULL" if not self.is_null else "",
        )

    def to_mermaid_er(self) -> str:
        type_str = self.type.replace(" ", "_")
        name = sanitize_mermaid(self.name, is_er=True)
        return f" {type_str} {name} {'PK' if self.is_key else ''}"

    def to_dot(self) -> str:
        base = config["DOT_ROW_TAGS"].format(
            ' ALIGN="LEFT" {port}',
            "{key_opening}{col_name}{key_closing} {type}{null}",
        )
        KO = config["DOT_KEY_OPENING"]
        KC = config["DOT_KEY_CLOSING"]
        return base.format(
            port=f'PORT="{self.name}"' if self.name else "",
            key_opening=KO if self.is_key else "",
            key_closing=KC if self.is_key else "",
            col_name=config["DOT_FONT_TAGS"].format(self.name),
            type=(
                config["DOT_FONT_TAGS"].format(f" [{self.type}]") if self.type is not None else ""
            ),
            null=" NOT NULL" if not self.is_null else "",
        )

    def to_puml(self) -> str:
        return " {} {} : {}{}{}".format(
            self.key_symbol,
            self.name,
            self.type.replace("(", "<").replace(")", ">"),
            " NOT NULL" if not self.is_null else "",
            "\n--" if self.is_key else "",
        )


class Relation(Drawable):
    """Represents a Relation in the intermediaty syntax."""

    RE = re.compile(
        r"""
        (?P<left_table>[^\s]+?)
        (?:\.\"(?P<left_column>.+)\")?
        \s*
        (?P<left_cardinality>[*?+1])
        --
        (?P<right_cardinality>[*?+1])
        \s*
        (?P<right_table>[^\s]+?)
        (?:\.\"(?P<right_column>.+)\")?
        \s*$
        """,
        re.VERBOSE,
    )
    cardinalities = {"*": "0..N", "?": "{0,1}", "+": "1..N", "1": "1", "": None}
    cardinalities_mermaid = {
        "*": "0..n",
        "?": "0..1",
        "+": "1..n",
    }
    cardinalities_crowfoot = {
        "*": "0+",
        "?": "one or zero",
        "+": "1+",
    }

    @staticmethod
    def make_from_match(match: re.Match) -> Relation:
        return Relation(**match.groupdict())

    def __init__(
        self,
        right_table,
        left_table,
        right_cardinality=None,
        left_cardinality=None,
        right_column=None,
        left_column=None,
    ):
        if (
            right_cardinality not in self.cardinalities.keys()
            or left_cardinality not in self.cardinalities.keys()
        ):
            raise ValueError(f"Cardinality should be in {self.cardinalities.keys()}")
        self.right_table = right_table
        self.right_column = right_column or ""
        self.left_table = left_table
        self.left_column = left_column or ""
        self.right_cardinality = right_cardinality
        self.left_cardinality = left_cardinality

    def to_markdown(self) -> str:
        return "{}{} {}--{} {}{}".format(
            self.left_table,
            "" if not self.left_column else f'."{self.left_column}"',
            self.left_cardinality,
            self.right_cardinality,
            self.right_table,
            "" if not self.right_column else f'."{self.right_column}"',
        )

    def to_mermaid(self) -> str:
        normalized = (
            Relation.cardinalities_mermaid.get(k, k)
            for k in (
                sanitize_mermaid(self.left_table),
                self.left_cardinality,
                self.right_cardinality,
                sanitize_mermaid(self.right_table),
            )
        )
        return '  {} "{}" -- "{}" {}'.format(*normalized)

    def to_mermaid_er(self) -> str:
        left = Relation.cardinalities_crowfoot.get(self.left_cardinality, self.left_cardinality)
        right = Relation.cardinalities_crowfoot.get(self.right_cardinality, self.right_cardinality)

        left_col = sanitize_mermaid(self.left_table, is_er=True)
        right_col = sanitize_mermaid(self.right_table, is_er=True)
        return f"{left_col} {left}--{right} {right_col} : has"

    def graphviz_cardinalities(self, card) -> str:
        if card == "":
            return ""
        return f"label=<<FONT>{self.cardinalities[card]}</FONT>>"

    def graphviz_crow_arrowheads(self, card):
        if card == "*":
            head = '="crowodot"'
        elif card == "?":
            head = '="teeodot"'
        elif card == "+":
            head = '="crowtee"'
        elif card == "1":
            head = '="teetee"'
        else:
            raise ValueError(f"unknown cardinality: {card}")
        return head

    def to_dot(self) -> str:
        if self.right_cardinality == self.left_cardinality == "":
            return ""
        cards = []
        edge = "--"
        if config["DOT_RELATION_GRAPH"] == "digraph":
            # digraph needs direction
            # https://graphviz.org/doc/info/lang.html#lexical-and-semantic-notes
            edge = "->"
        if config["DOT_RELATION_STYLE"] == "crow":
            if self.right_cardinality and self.left_cardinality:
                cards.append('dir="both"')
        if self.left_cardinality != "":
            if config["DOT_RELATION_STYLE"] == "crow":
                cards.append("arrowhead" + self.graphviz_crow_arrowheads(self.left_cardinality))
            else:
                cards.append("tail" + self.graphviz_cardinalities(self.left_cardinality))
        if self.right_cardinality != "":
            if config["DOT_RELATION_STYLE"] == "crow":
                cards.append("arrowtail" + self.graphviz_crow_arrowheads(self.right_cardinality))
            else:
                cards.append("head" + self.graphviz_cardinalities(self.right_cardinality))
        left_col = f':"{self.left_column}"' if self.left_column else ""
        right_col = f':"{self.right_column}"' if self.right_column else ""
        return f'"{self.left_table}"{left_col} {edge} "{self.right_table}"{right_col} [{",".join(cards)}];'

    def to_puml(self) -> str:
        __puml_left_cardinalities = {
            "*": "}o",
            "?": "|o",
            "+": "}1",
            "1": "||",
            "": None,
        }
        __puml_right_cardinalities = {
            "*": "o{",
            "?": "o|",
            "+": "1{",
            "1": "||",
            "": None,
        }
        left_col = f"::{self.left_column}" if self.left_column else ""
        right_col = f"::{self.right_column}" if self.right_column else ""

        return (
            f"{self.left_table}{left_col}"
            f" {__puml_left_cardinalities[self.left_cardinality]}"
            f"--{__puml_right_cardinalities[self.right_cardinality]}"
            f" {self.right_table}{right_col}"
        )

    def __eq__(self, other: object) -> bool:
        if super().__eq__(other):
            return True
        if not isinstance(other, Relation):
            return False
        other_inversed = Relation(
            right_table=other.left_table,
            right_column=other.left_column,
            left_table=other.right_table,
            left_column=other.right_column,
            right_cardinality=other.left_cardinality,
            left_cardinality=other.right_cardinality,
        )
        return other_inversed.__dict__ == self.__dict__


class Table(Drawable):
    """Represents a Table in the intermediaty syntax."""

    RE = re.compile(r"\[(?P<name>[^]]+)\]")

    def __init__(self, name: str, columns: list[Column]) -> None:
        self.name = name
        self.columns = columns

    @staticmethod
    def make_from_match(match: re.Match) -> Table:
        return Table(name=match.group("name"), columns=[])

    @property
    def header_markdown(self) -> str:
        return f"[{self.name}]"

    def to_markdown(self) -> str:
        return self.header_markdown + "\n" + "\n".join(c.to_markdown() for c in self.columns)

    def to_mermaid(self) -> str:
        columns = [c.to_mermaid() for c in self.columns]
        name = sanitize_mermaid(self.name)
        return f"  class {name}{{\n   " + "\n   ".join(columns) + "\n  }"

    def to_mermaid_er(self) -> str:
        columns = [c.to_mermaid_er() for c in self.columns]
        name = sanitize_mermaid(self.name, is_er=True)
        return f"{name} {{\n" + "\n  ".join(columns) + "\n}"

    @property
    def columns_sorted(self):
        return sorted(self.columns, key=operator.attrgetter("name"))

    @property
    def header_dot(self) -> str:
        return config["DOT_ROW_TAGS"].format("", f'<B><FONT POINT-SIZE="16">{self.name}</FONT></B>')

    def to_dot(self) -> str:
        body = "".join(c.to_dot() for c in self.columns)
        return config["DOT_TABLE"].format(self.name, self.header_dot, body)

    def to_puml(self) -> str:
        columns = [c.to_puml() for c in self.columns]
        name = sanitize_mermaid(self.name)
        return f"entity  {name}{{\n" + "\n  ".join(columns) + "\n}"

    def __str__(self) -> str:
        return self.header_markdown

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Table):
            return False
        if other.name != self.name:
            return False

        if self.columns_sorted != other.columns_sorted:
            return False
        return True



from eralchemy import render_er
from eralchemy.cst import config, dot_star_primary, dot_top_down, reset_config


def render_dot():
    reset_config()
    assert config["DOT_KEY_OPENING"] == "<u>"
    assert config["DOT_KEY_CLOSING"] == "</u>"
    assert "rankdir=LR" in config["DOT_GRAPH_BEGINNING"]
    output = render_er("../example/forum.er", output=None, mode="dot").decode("utf-8")
    assert "<u><FONT>id</FONT></u> " in output

    with open("expected_plain_output.dot") as f:
        expected = f.read()
    assert expected == output


def render_dot_with_config():
    reset_config()
    dot_star_primary()
    assert config["DOT_KEY_OPENING"] == "*"
    assert config["DOT_KEY_CLOSING"] == ""
    output = render_er("../example/forum.er", output=None, mode="dot").decode("utf-8")
    assert "*<FONT>id</FONT> " in output

    with open("expected_star_primary_output.dot") as f:
        expected = f.read()
    assert expected == output


def reder_dot_top_down():
    reset_config()
    dot_top_down()
    assert "rankdir=TB" in config["DOT_GRAPH_BEGINNING"]
    output = render_er("../example/forum.er", output=None, mode="dot").decode("utf-8")
    assert "rankdir=TB" in output



import pytest

from eralchemy.main import (
    all_to_intermediary,
    filter_resources,
    get_output_mode,
    intermediary_to_dot,
    intermediary_to_markdown,
    intermediary_to_mermaid,
    intermediary_to_mermaid_er,
    intermediary_to_schema,
)
from tests.common import (
    Base,
    check_filter,
    check_intermediary_representation_simple_all_table,
    check_intermediary_representation_simple_table,
    check_tables_columns,
    check_tables_relationships,
    markdown,
    relationships,
    tables,
)


def test_all_to_intermediary_base():
    tables, relationships = all_to_intermediary(Base)
    check_intermediary_representation_simple_all_table(tables, relationships)


def test_all_to_intermediary_db_sqlite(sqlite_db_uri):
    tables, relationships = all_to_intermediary(sqlite_db_uri)
    check_intermediary_representation_simple_table(tables, relationships)


@pytest.mark.external_db
def test_all_to_intermediary_db(pg_db_uri):
    tables, relationships = all_to_intermediary(pg_db_uri)
    check_intermediary_representation_simple_table(tables, relationships)


def test_all_to_intermediary_markdown():
    tables, relationships = all_to_intermediary(markdown.split("\n"))
    check_intermediary_representation_simple_table(tables, relationships)


def test_all_to_intermediary_fails():
    with pytest.raises(ValueError):
        all_to_intermediary("plop")


def test_filter_no_include_no_exclude():
    actual_tables, actual_relationships = filter_resources(tables, relationships)
    check_filter(actual_tables, actual_relationships)


@pytest.mark.parametrize(
    "include_tables",
    (
        ["parent", "child", "excl"],
        ["^(?!exc)\\w+$"],  # all not starting with excl
        ["parent", "child"],
        ["parent", "^ch.*"],
        ["par.*", "child"],
    ),
)
def test_filter_include_tables(include_tables):
    actual_tables, actual_relationships = filter_resources(
        tables,
        relationships,
        include_tables=include_tables,
    )
    check_tables_relationships(actual_tables, actual_relationships)


@pytest.mark.parametrize(
    "exclude_tables",
    (
        ["ent", "ld", ".*lude"],
        ["par", "ch", "^exclude"],
        ["ar.*", "hild", "exc.*"],
        ["exclude"],
    ),
)
def test_filter_exclude_tables(exclude_tables):
    actual_tables, actual_relationships = filter_resources(
        tables,
        relationships,
        exclude_tables=exclude_tables,
    )
    check_tables_relationships(actual_tables, actual_relationships)


@pytest.mark.parametrize(
    "include_columns",
    (
        ["id"],
        ["id.*", "not_match"],
    ),
)
def test_filter_include_columns(include_columns):
    actual_tables, actual_relationships = filter_resources(
        tables,
        relationships,
        include_columns=include_columns,
    )
    check_tables_columns(actual_tables, id_is_included=True)


@pytest.mark.parametrize(
    "exclude_columns",
    (
        ["id"],
        ["i."],
    ),
)
def test_filter_exclude_columns(exclude_columns):
    actual_tables, actual_relationships = filter_resources(
        tables,
        relationships,
        exclude_columns=exclude_columns,
    )
    check_tables_columns(actual_tables, id_is_included=False)


def test_get_output_mode():
    # access .func for partial
    assert get_output_mode("hello.png", "auto").func == intermediary_to_schema
    assert get_output_mode("hello.er", "auto") == intermediary_to_markdown
    assert get_output_mode("hello.dot", "auto") == intermediary_to_dot
    assert get_output_mode("hello.md", "auto") == intermediary_to_mermaid

    assert get_output_mode("anything", "graph") == intermediary_to_schema
    assert get_output_mode("anything", "dot") == intermediary_to_dot
    assert get_output_mode("anything", "er") == intermediary_to_markdown
    assert get_output_mode("anything", "mermaid") == intermediary_to_mermaid
    assert get_output_mode("anything", "mermaid_er") == intermediary_to_mermaid_er

    with pytest.raises(ValueError):
        get_output_mode("anything", "mode")


def test_import_render_er():
    from eralchemy import render_er  # noqa: F401



import re

from eralchemy.main import _intermediary_to_markdown
from tests.common import (
    child,
    child_id,
    child_parent_id,
    parent,
    parent_id,
    parent_name,
    relation,
)

column_re = re.compile(r'(?P<key>\*?)(?P<name>[^*].+) \{label:"(?P<type>.+)"\}')


def test_all_to_er():
    tables = [child, parent]
    relations = [relation]
    output = _intermediary_to_markdown(tables, relations)
    for element in relations + tables:
        assert element.to_markdown() in output


def assert_column_well_rendered_to_er(col):
    col_er = col.to_markdown().strip()
    col_parsed = column_re.match(col_er)
    assert col_parsed.group("key") == ("*" if col.is_key else "")
    assert col_parsed.group("name") == col.name
    assert col_parsed.group("type") == col.type


def test_column_to_er():
    assert_column_well_rendered_to_er(parent_id)
    assert_column_well_rendered_to_er(parent_name)
    assert_column_well_rendered_to_er(child_id)
    assert_column_well_rendered_to_er(child_parent_id)


def test_relation():
    assert relation.to_markdown() in [
        'parent."id" *--? child."parent_id"',
        'child."parent_id" *--? parent."id"',
    ]


def assert_table_well_rendered_to_er(table):
    assert table.header_markdown == "[" + table.name + "]"
    table_er = table.to_markdown()
    for col in table.columns:
        assert col.to_markdown() in table_er


def test_table():
    assert_table_well_rendered_to_er(child)
    assert_table_well_rendered_to_er(parent)



import pytest

from eralchemy.main import _intermediary_to_markdown
from eralchemy.models import Column, Relation, Table
from eralchemy.parser import (
    DuplicateColumnException,
    DuplicateTableException,
    NoCurrentTableException,
    ParsingException,
    RelationNoColException,
    line_iterator_to_intermediary,
    parse_line,
    remove_comments_from_line,
    update_models,
)
from tests import common as c

# examples from https://github.com/BurntSushi/erd/blob/master/examples/nfldb.er
table_lst = [
    "[player]",
    "[team]",
    "[game]",
    "[drive]",
]

relations_lst = [
    "player      *--1 team",
    "game        *--1 team",
    "game        *--* team",
    "drive       1--1 team",
    "play        ?--1 team",
    "play_player *--+ team",
]

columns_lst = [
    # '*+gsis_id', # TODO add fk
    # '*+drive_id',
    "*play_id",
    "time",
    "pos_team",
    "yardline",
    "down",
    "yards_to_go",
]
elements_lst = table_lst + relations_lst + columns_lst


def test_remove_from_lines():
    r = remove_comments_from_line
    for code in elements_lst:
        assert r(code) == code
        assert r(f"{code} ## some comment") == code
        assert r(f"   {code}") == code
        assert r(f"{code}   ") == code
        assert r(f"{code} ## some comment") == code
        assert r(f"{code} #  # some comment") == code
        assert r(f"   {code} #  # some comment") == code
        assert r(f"{code}") == code
        assert r(f"#{code} #  # some comment") == ""
        assert r(f"# #{code} #  # some comment") == ""
        assert r(f"##{code}") == ""


def test_parse_line_type():
    col = parse_line('parent_id {label:"INTEGER"}')
    assert col.type == "INTEGER"


def test_parse_line():
    for s in columns_lst:
        rv = parse_line(s)
        assert rv.name == s.replace("*", "")
        assert isinstance(rv, Column)

    for s in relations_lst:
        rv = parse_line(s)  # type: Relation
        assert rv.right_table == s[16:].strip()
        assert rv.right_column == ""
        assert rv.left_table == s[:12].strip()
        assert rv.left_column == ""
        assert rv.right_cardinality == s[15]
        assert rv.left_cardinality == s[12]
        assert isinstance(rv, Relation)

    for s in table_lst:
        rv = parse_line(s)
        assert rv.name == s[1:-1]
        assert rv.columns == []
        assert isinstance(rv, Table)


def test_update_models_fails_no_current_table():
    for new_obj in (c.relation, c.parent_id):
        with pytest.raises(NoCurrentTableException):
            update_models(new_obj, None, [], [])


def test_update_models_fails_relation_no_col():
    with pytest.raises(RelationNoColException):
        update_models(
            new_obj=c.relation,
            current_table=c.parent,
            tables=[c.parent],
            relations=[],
        )


def test_update_models_fails_duplicate_col():
    with pytest.raises(DuplicateColumnException):
        update_models(
            new_obj=c.parent_name,
            current_table=c.parent,
            tables=[c.parent],
            relations=[],
        )


def test_update_models_fails_duplicate_table():
    with pytest.raises(DuplicateTableException):
        update_models(
            new_obj=c.parent,
            current_table=c.parent,
            tables=[c.parent],
            relations=[],
        )


def test_update_models_add_relation():
    current_table, tables, relations = update_models(
        new_obj=c.relation,
        current_table=c.parent,
        tables=[c.parent, c.child],
        relations=[],
    )
    assert c.relation in relations


def test_update_models_add_table():
    current_table, tables, relations = update_models(
        new_obj=c.child,
        current_table=c.parent,
        tables=[
            c.parent,
        ],
        relations=[],
    )
    assert c.child in tables


def test_update_models_new_obj_bad_class():
    with pytest.raises(ValueError):
        update_models(
            new_obj=c.Child,
            current_table=c.parent,
            tables=[
                c.parent,
            ],
            relations=[],
        )


def test_update_models_add_column():
    parent = Table(
        name="parent",
        columns=[c.parent_id],
    )
    current_table, tables, relations = update_models(
        new_obj=c.parent_name,
        current_table=parent,
        tables=[
            parent,
        ],
        relations=[],
    )
    assert c.parent_name in current_table.columns
    assert c.parent_name in tables[0].columns


def test_integration_parser():
    tables, relations = line_iterator_to_intermediary(c.markdown.split("\n"))
    c.assert_lst_equal(tables, c.tables)
    c.assert_lst_equal(relations, [c.relation, c.exclude_relation])


def test_generate_and_parse():
    markdown = _intermediary_to_markdown(c.tables, [c.relation])
    tables, relations = line_iterator_to_intermediary(markdown.split("\n"))
    c.assert_lst_equal(tables, c.tables)
    c.assert_lst_equal(relations, [c.relation])


def test_integration_errors():
    markdown_broken = """
            name {label:"VARCHAR(255)"}
        [child]
            *id {label:"INTEGER"}
            parent_id {label:"INTEGER"}
        parent *--? child
        """
    with pytest.raises(ParsingException):
        line_iterator_to_intermediary(markdown_broken.split("\n"))
    # TODO check error



import re

import pytest

from eralchemy.sqla import (
    column_to_intermediary,
    database_to_intermediary,
    declarative_to_intermediary,
    table_to_intermediary,
)
from tests.common import (
    Base,
    Child,
    Parent,
    Relation,
    Table,
    check_intermediary_representation_simple_all_table,
    check_intermediary_representation_simple_table,
    child,
    child_id,
    child_parent_id,
    exclude_relation,
    parent,
    parent_id,
    parent_name,
    relation,
)


def check_column(column, column_intermediary):
    output = column_to_intermediary(column)
    assert output.name == column_intermediary.name
    assert output.is_key is column_intermediary.is_key
    assert output.type == column_intermediary.type


def test_columns_parent():
    check_column(column=Parent.id, column_intermediary=parent_id)

    check_column(column=Parent.name, column_intermediary=parent_name)


def test_columns_child():
    check_column(column=Child.id, column_intermediary=child_id)

    check_column(column=Child.parent_id, column_intermediary=child_parent_id)


def test_declarative_to_intermediary():
    tables, relationships = declarative_to_intermediary(Base)
    check_intermediary_representation_simple_all_table(tables, relationships)


def table_equals_helper(sqla_table, expected_table):
    table = table_to_intermediary(sqla_table.__table__)
    assert len(table.columns) == len(expected_table.columns)
    assert table.name == expected_table.name
    for col in table.columns:
        assert col in expected_table.columns


def test_tables():
    table_equals_helper(Child, child)
    table_equals_helper(Parent, parent)


@pytest.mark.external_db
def test_database_to_intermediary(pg_db_uri):
    tables, relationships = database_to_intermediary(pg_db_uri)
    check_intermediary_representation_simple_table(tables, relationships)


@pytest.mark.external_db
def test_database_to_intermediary_with_schema(pg_db_uri):
    tables, relationships = database_to_intermediary(pg_db_uri, schema="eralchemy_test")

    assert len(tables) == 3
    assert len(relationships) == 2
    assert all(isinstance(t, Table) for t in tables)
    assert all(isinstance(r, Relation) for r in relationships)
    # Not in because different schema.
    assert relation not in relationships
    assert exclude_relation not in relationships


@pytest.mark.external_db
def test_database_to_intermediary_with_multiple_schemas(pg_db_uri):
    tables, relationships = database_to_intermediary(
        pg_db_uri,
        schema="public, eralchemy_test",
    )

    assert len(tables) == 6
    assert len(relationships) == 4
    assert all(isinstance(t, Table) for t in tables)
    assert all(isinstance(r, Relation) for r in relationships)
    # Not in because different schema.
    assert relation not in relationships
    assert exclude_relation not in relationships


def test_flask_sqlalchemy():
    from flask import Flask
    from flask_sqlalchemy import SQLAlchemy

    from eralchemy.main import all_to_intermediary

    app = Flask(__name__)
    app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:////tmp/test.db"
    app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = True
    db = SQLAlchemy(app)
    model = db.Model
    model.metadata = Base.metadata
    tables, relationships = all_to_intermediary(db.Model)
    check_intermediary_representation_simple_all_table(tables, relationships)


@pytest.mark.external_db
def test_table_names_in_relationships(pg_db_uri):
    tables, relationships = database_to_intermediary(pg_db_uri)
    table_names = [t.name for t in tables]

    # Assert column names are table names
    assert all(r.right_table in table_names for r in relationships)
    assert all(r.left_table in table_names for r in relationships)

    # Assert column names match table names
    for r in relationships:
        r_name = table_names[table_names.index(r.right_table)]
        l_name = table_names[table_names.index(r.left_table)]

        # Table name in relationship should *NOT* have a schema
        assert r_name.find(".") == -1
        assert l_name.find(".") == -1


@pytest.mark.external_db
def test_table_names_in_relationships_with_schema(pg_db_uri):
    schema_name = "test"
    matcher = re.compile(rf"{schema_name}\.[\S+]", re.I)
    tables, relationships = database_to_intermediary(pg_db_uri, schema=schema_name)
    table_names = [t.name for t in tables]

    # Assert column names match table names, including schema
    assert all(r.right_table in table_names for r in relationships)
    assert all(r.left_table in table_names for r in relationships)

    # Assert column names match table names, including schema
    for r in relationships:
        r_name = table_names[table_names.index(r.right_col)]
        l_name = table_names[table_names.index(r.left_col)]

        # Table name in relationship *SHOULD* have a schema
        assert re.match(matcher, r_name) is not None
        assert re.match(matcher, l_name) is not None



from eralchemy.helpers import check_args
from eralchemy.main import get_argparser


def parse_test(lst_arguments):
    parser = get_argparser()
    args = parser.parse_args(lst_arguments)
    check_args(args)


def test_version():
    parse_test(["-v"])


def test_normal():
    parse_test("-i sqlite:///relative/path/to/db.db -o erd_from_sqlite.pdf".split(" "))



"""All the constants used in the module."""

DOT_TABLE = (
    '"{}" [label=<<FONT FACE="Helvetica"><TABLE BORDER="0" CELLBORDER="1"'
    ' CELLPADDING="4" CELLSPACING="0">{}{}</TABLE></FONT>>];'
)
DOT_FONT_TAGS = "<FONT>{}</FONT>"
# Used for each row in the table.
DOT_ROW_TAGS = "<TR><TD{}>{}</TD></TR>"
DOT_GRAPH_BEGINNING = """graph {
    graph [rankdir=LR];
    node [label="\\N",
        shape=plaintext
    ];
    edge [color=gray50,
        minlen=2,
        style=dashed
    ];"""
MARKDOWN_TITLE = 'title {{label: "{}", size: "40"}}'

DEFAULT_CONFIG = {
    "DOT_TABLE": DOT_TABLE,
    "DOT_FONT_TAGS": DOT_FONT_TAGS,
    "DOT_ROW_TAGS": DOT_ROW_TAGS,
    "DOT_GRAPH_BEGINNING": DOT_GRAPH_BEGINNING,
    "MARKDOWN_TITLE": MARKDOWN_TITLE,
    "DOT_KEY_OPENING": "<u>",
    "DOT_KEY_CLOSING": "</u>",
    "DOT_RELATION_GRAPH": "graph",
    "DOT_RELATION_STYLE": "",
}

config = DEFAULT_CONFIG.copy()


def dot_star_primary():
    config["DOT_KEY_OPENING"] = "*"
    config["DOT_KEY_CLOSING"] = ""


def dot_star_underline():
    config["DOT_KEY_OPENING"] = "<u>"
    config["DOT_KEY_CLOSING"] = "</u>"


def dot_top_down():
    config["DOT_GRAPH_BEGINNING"] = config["DOT_GRAPH_BEGINNING"].replace(
        "rankdir=LR", "rankdir=TB"
    )


def dot_left_right():
    config["DOT_GRAPH_BEGINNING"] = config["DOT_GRAPH_BEGINNING"].replace(
        "rankdir=TB", "rankdir=LR"
    )


def dot_digraph():
    config["DOT_GRAPH_BEGINNING"] = config["DOT_GRAPH_BEGINNING"].replace("graph {", "digraph {")
    config["DOT_RELATION_GRAPH"] = "digraph"


def dot_crowfoot():
    config["DOT_RELATION_STYLE"] = "crow"


def reset_config():
    for key, value in DEFAULT_CONFIG.items():
        config[key] = value



import re
import sys
from multiprocessing import Process

import pytest
from pygraphviz import AGraph

from eralchemy.main import _intermediary_to_dot
from tests.common import (
    child,
    child_id,
    child_parent_id,
    parent,
    parent_id,
    parent_name,
    relation,
)

column_re = re.compile(r"\<TR\>\<TD\ ALIGN\=\"LEFT\"\ PORT\=\".+\">(.*)\<\/TD\>\<\/TR\>")
header_re = re.compile(
    r"\<TR\>\<TD\>\<B\>\<FONT\ POINT\-SIZE\=\"16\"\>(.*)" r"\<\/FONT\>\<\/B\>\<\/TD\>\<\/TR\>"
)
column_inside = re.compile(
    r"(?P<key_opening>.*)\<FONT\>(?P<name>.*)\<\/FONT\>"
    r"(?P<key_closing>.*)\<FONT\>\ \[(?P<type>.*)\]\<\/FONT\>"
)


# This test needs fixing with move to graphviz
def assert_is_dot_format(dot):
    """Checks that the dot is usable by graphviz."""

    # We launch a process calling graphviz to render the dot. If the exit code is not 0 we assume that the syntax
    # wasn't good
    def run_graph(dot):
        """Runs graphviz to see if the syntax is good."""
        graph = AGraph()
        graph = graph.from_string(dot)
        extension = "png"
        graph.draw(path="output.png", prog="dot", format=extension)
        sys.exit(0)

    p = Process(target=run_graph, args=(dot,))
    p.start()
    p.join()
    assert p.exitcode == 0


def test_all_to_dot():
    tables = [child, parent]
    relations = [relation]
    output = _intermediary_to_dot(tables, relations)
    assert_is_dot_format(output)
    for element in relations + tables:
        assert element.to_dot() in output


@pytest.mark.parametrize(
    "title",
    ("Test Title", "häßlicher Titel!", "<div> -not.parsed_ </div>"),  # codespell:ignore
)
def test_all_to_dot_with_title(title):
    tables = [child, parent]
    relations = [relation]
    output = _intermediary_to_dot(tables, relations, title=title)
    assert_is_dot_format(output)
    for element in relations + tables:
        assert element.to_dot() in output


def assert_column_well_rendered_to_dot(col):
    col_no_table = column_re.match(col.to_dot()).groups()
    assert len(col_no_table) == 1
    col_parsed = column_inside.match(col_no_table[0])
    assert col_parsed.group("key_opening") == ("<u>" if col.is_key else "")
    assert col_parsed.group("name") == col.name
    assert col_parsed.group("key_closing") == ("</u> " if col.is_key else " ")
    assert col_parsed.group("type") == col.type


def test_column_is_dot_format():
    assert_column_well_rendered_to_dot(parent_id)
    assert_column_well_rendered_to_dot(parent_name)
    assert_column_well_rendered_to_dot(child_id)
    assert_column_well_rendered_to_dot(child_parent_id)


def test_relation():
    relation_re = re.compile(
        r"\"(?P<l_table>.+)\":\"(?P<l_column>.+)\"\ \-\-\ \"(?P<r_table>.+)\":\"(?P<r_column>.+)\"\ "
        r"\[taillabel\=\<\<FONT\>(?P<l_card>.+)\<\/FONT\>\>"
        r"\,headlabel\=\<\<FONT\>(?P<r_card>.+)\<\/FONT\>\>\]\;"
    )
    dot = relation.to_dot()
    r = relation_re.match(dot)
    assert r.group("l_table") == "child"
    assert r.group("l_column") == "parent_id"
    assert r.group("r_table") == "parent"
    assert r.group("r_column") == "id"
    assert r.group("l_card") == "0..N"
    assert r.group("r_card") == "{0,1}"


def assert_table_well_rendered_to_dot(table):
    matches = header_re.match(table.header_dot).groups()
    assert len(matches) == 1
    assert matches[0] == table.name
    table_dot = table.to_dot()
    for col in table.columns:
        assert col.to_dot() in table_dot


def test_table():
    assert_table_well_rendered_to_dot(child)
    assert_table_well_rendered_to_dot(parent)
